# This include the high-dimensional Knockoff and Knockoff plus fliter to computer FDR and Power
# This master R file includes all the R functions for implementing the knockoff filter and 
# knockoff filter plus in high-dimensional linear regression.
#
# The IPAD framework has a range of potential applications involving precision 
# (innovated data) matrix including:
#   Linear and nonlinear classifications;
#   Dimension reduction; 
#   Portfolio management;
#   Feature screening and simultaneous inference.
#
# Written by 
# Yingying Fan (fanyingy@marshall.usc.edu) and Jinchi Lv (jinchilv@marshall.usc.edu), Mahrad Sharifvaghefi (sharifva@usc.edu), Yoshimasa Uematsu (yoshimasa.uematsu.e7@tohoku.ac.jp)
# USC, Nov. 2017
#
#
# List of R functions for ISEE:
# Part I: main functions
#   1) hdknockoff: main function of high-dimensional knockoff filter
#   2) hdknockoff_plus: main function of high-dimensional knockoff+
#   3) hdknockoffX: function generating the knockoff design matrix
# Part II: additional functions
#   1) fdr: calculating fdr
#   2) pow: calculating power
#

###########################################################################
# Part I: main functions
# Part I 1) hdknockoff: high-dimensional knockoff filter
# Input:
#   X:   n x p design matrix
#   y:   n x 1 response vector
#   q:   nominal FDR level
# Output: 
#   S:   set of discovered variables

hdknockoff <- function(X, y, q, lambda = "CV"){
  
  n = nrow(X)
  p = ncol(X)/2
  
  # generate the knockoff design matrix

  if (lambda == "CV") {
    # Finding Optimal value of Lambda
    Op_lambda <- cv.glmnet(X,y)$lambda.min
    # apply glmnet to find beta vector under original scale
    Beta <- glmnet(X,y,lambda = Op_lambda)$beta
    Beta <- as.vector(Beta)
    # beta vector under aligned scale
    stan_X = scale(X, center=FALSE, scale = TRUE)/sqrt(n-1)
    y.norm = attr(stan_X,"scale")*sqrt(n-1)
    betap.stan = Beta*y.norm
  }
  
  # Calculate knockoff test statistics W_j
  W = abs(betap.stan[1:p]) - abs(betap.stan[(p+1):(2*p)])

  
  # find the knockoff threshold T
  t = sort(c(0, abs(W)))
  ratio = c(rep(0, p))
  for (j in 1:p) {
    ratio[j] = sum(W <= -t[j])/max(1, sum(W >= t[j]))
  }
  id = which(ratio <= q)[1]
  if(length(id) == 0){
    T = Inf
  } else {
    T = t[id]
  }
  
  # set of discovered variables
  S = which(W >= T)
  
  return(S)
}
###########################################################################

###########################################################################
# Part I: main functions
# Part I 1) hdknockoff plus: high-dimensional knockoff plus filter
# Input:
#   X:   n x p design matrix
#   y:   n x 1 response vector
#   q:   nominal FDR level
# Output: 
#   S_plus:   set of discovered variables

hdknockoff_plus <- function(X, y, q, lambda = "CV"){
  
  n = nrow(X)
  p = ncol(X)/2
  
  # generate the knockoff design matrix
  if (lambda == "CV") {
    # Finding Optimal value of Lambda
    Op_lambda <- cv.glmnet(X,y)$lambda.min
    # apply glmnet to find beta vector under original scale
    Beta <- glmnet(X,y,lambda = Op_lambda)$beta
    Beta <- as.vector(Beta)
    # beta vector under aligned scale
    stan_X = scale(X, center=F, scale=T)/sqrt(n-1)
    y.norm = attr(stan_X,"scale")*sqrt(n-1)
    betap.stan = Beta*y.norm
  }
  
  # Calculate knockoff test statistics W_j
  W = abs(betap.stan[1:p]) - abs(betap.stan[(p+1):(2*p)])
  
  # find the knockoff threshold T
  t = sort(c(0, abs(W)))
  ratio = c(rep(0, p))
  for (j in 1:p) {
    ratio[j] = (1+sum(W <= -t[j]))/max(1, sum(W >= t[j]))
  }
  id = which(ratio <= q)[1]
  if(length(id) == 0){
    T = Inf
  } else {
    T = t[id]
  }
  
  # set of discovered variables
  S_plus = which(W >= T)
  
  return(S_plus)
}
###########################################################################


###########################################################################
# Part I: main functions
# Part I 2) hdknockoffX: function generating the knockoff design matrix
# Input:
#   X:      n x p design matrix
#   y:      n x 1 response vector
# Output: 
#   Xnew:   n x (2p) design matrix = [X X_ko]
# Methods: 1- facpermu: Knockoff variables are constructed by using factor model with derangement premutation of residuals
#          2- facrand: Knockoff variables are constructed by using factor model with random error drown form unit sphere 
#          3- linsparse: Knockoff variables are constructed by using linear sparse model
#          4- BCknockoff1: Barber and Candes equicorrelated method for construction of knockoff variables
#          5- hdBCknockoff1: Barber and Candes equicorrelated method for construction of knockoff variables in high dimensional setting
hdknockoffX <- function(X,y,r,beta.true,method = "facrand", lambda = "CV"){
  if (method == "facrand") {
    n = nrow(X)
    p = ncol(X)
    C = fFact(X,r)
    res = X - C
    resScale = sqrt(apply(res^2,2,sum))
    resScale = diag(resScale)
    e = matrix(rnorm(n*p),n,p)
    eScale = sqrt(apply(e^2,2,sum))
    eScale = diag(eScale^-1)
    #mean_res = as.matrix(apply(res,2,mean))
    #sd_res = diag(apply(res,2,sd))
    nres = e %*% resScale %*% eScale # %*% sd_res + matrix(1,n,1) %*% t(mean_res)
    X_ko = C + nres
    obj = cbind(X, X_ko)
  } else if (method == "BCknockoff1") {
    X_ko <- knockoff.create(X,method = "equicorrelated")
    obj = cbind(X, X_ko)
  } else if (method == "hdBCknockoff1" || method == "hdBCknockoff2") {
    n = nrow(X)
    #### splite the sample 
    SSample = sample.Splite(X,y)
    X0      = SSample$X0
    y0      = SSample$y0
    X1      = SSample$X1
    y1      = SSample$y1
    #################################################
    ### using the data to do screening and get a selected set
    Obj.scr.set = sel.set(X0, y0, n)
    scr.set     = Obj.scr.set$scr.set
    ## judge whether the selected model including the true model
    #cv.id[ii] <- as.numeric(all(id.true %in% scr.set))
    
    ################################################
    ## based on selected set to get the reduced model and select the 
    ## variable using knockoff and knockoff plus respectively
    sel.v   = sel.variable(scr.set, X0, X1, beta.true)
    X1.S    = sel.v$X1.sel
    X0.S    = sel.v$X0.sel
    beta.S  = sel.v$beta.sel
    normc = function(X) t(t(X)/sqrt(colSums(X^2)))
    X0.S  = normc(X0.S)
    X1.S  = normc(X1.S)
    X_ko = knockoff.create(X1.S, method = "equicorrelated")
    if (method == "hdBCknockoff1") {
      Xnew <- cbind(X1.S, X_ko)
      Supp = hdknockoff(Xnew, y1, q,lambda = "CV")
      fdp = fdr(Supp, beta.S)
      tdp = pow(Supp, beta.S)
      Supp_plus = hdknockoff_plus(Xnew, y1, q,lambda = "CV")
      fdp_plus = fdr(Supp_plus, beta.S)
      tdp_plus = pow(Supp_plus, beta.S)
      obj  = list(fdp = fdp, tdp = tdp, fdp_plus = fdp_plus, tdp_plus = tdp_plus )
    }else if (method == "hdBCknockoff2") {
      ## Next we use the recycling data to compute the FDR and Power
      X_S   = rbind(X0.S,X1.S)
      X_KO  = rbind(X0.S,X_ko)
      y_S   = c(y0,y1)
      Xnew <- cbind(X_S, X_KO)
      Supp = hdknockoff(Xnew, y_S, q,lambda = "CV")
      fdp = fdr(Supp, beta.S)
      tdp = pow(Supp, beta.S)
      Supp_plus = hdknockoff_plus(Xnew, y_S, q,lambda = "CV")
      fdp_plus = fdr(Supp_plus, beta.S)
      tdp_plus = pow(Supp_plus, beta.S)
      obj  = list(fdp = fdp, tdp = tdp, fdp_plus = fdp_plus, tdp_plus = tdp_plus)
    }
  }
  return(obj)
}
###########################################################################


###########################################################################
# Part I: main functions
# Part I 3) slassotf: function implementing tuning-free sparse regression with scaled Lasso
# Windows version; Mac or Linux version
# Lasso implemented with coordinate descent

slassotf <- function(X, y, regfactor = "log", sis.use = 0, maxite = 50, tol = 1e-2){
  # detect the OS
  if (Sys.info()['sysname'] == 'Windows') {
    windows.use = 1
  }else{
    windows.use = 0}
  
  n = nrow(X)
  p = ncol(X)
  
  if (regfactor == "log") {
    reg.fac = log(n)
  }
  if (regfactor == "log_p") {
    reg.fac = log(p)
  }
  if (regfactor == "one") {
    reg.fac = 1
  }
  if (regfactor == "sqrt") {
    reg.fac = sqrt(n)
  }
  
  if (sis.use) {
    set0<-order(abs(cor(y, X)),decreasing=T)
    sis.size = min(floor(n/log(n)),p)
    sis.set= set0[1:sis.size]
  }else {
    sis.set = c(1:p)
  }
  
  # rescale each column of n x p design matrix X to have unit L2-norm
  stan_X = scale(X, center=F, scale=T)/sqrt(n-1)
  y.norm = attr(stan_X,"scale")*sqrt(n-1) 
  X = stan_X
  
  # initialization of lambda
  gata2 <- sqrt(n)/(log(p)*2*p)
  B2 <- qt(1-gata2, n-1)
  lambda <- B2/sqrt(n-1+B2^2)
  
  sigma.old <- 1
  beta.old <- c(rep(0, length(sis.set)))
  last_beta_diff <- 0
  inner <- 0
  while (inner < 5) {
    reg <- sigma.old*lambda*reg.fac
    beta <- slasso(as.matrix(X[, sis.set]), as.vector(y), reg, c(rep(0, length(sis.set))), windows.use, maxite, tol)
    if (reg.fac == 1) {
      beta = beta*(abs(beta) > 1*lambda)
    }else{
      beta = beta*(abs(beta) > 0.5*lambda)
    }
    
    sel.mod = which(beta != 0)
    beta[sel.mod] = slasso(as.matrix(X[, sis.set[sel.mod]]), as.vector(y), reg, c(rep(0, length(sel.mod))), windows.use, maxite, tol)
    beta_diff <- max(abs(beta - beta.old))
    A <- y - X[, sis.set] %*% beta
    sigma <- sqrt(sum(A^2))/sqrt(max(n-sum(beta!=0),0)+1e-20)
    sigma_diff <- abs(sigma - sigma.old)
    
    if (sigma_diff < tol) 
      break
    else if (sigma_diff < tol * 0.1& 
             abs(beta_diff - last_beta_diff) < tol) 
      break
    else {
      sigma.old <- sigma
      beta.old <- beta
      last_beta_diff <- beta_diff
    }
    inner <- inner + 1
  }
  
  # beta vector under aligned scale
  betap.stan = c(rep(0, p))
  betap.stan[sis.set] <- as.vector(beta)
  # beta vector under original scale
  betap = scale(t(betap.stan), center=F, scale=y.norm)
  
  obj=list(betap = betap, betap.stan = betap.stan)
  return(obj)
}
###########################################################################
fFact <- function(X, r){
  # SX = scale(X)
  # mXX = SX %*%t(SX)
  # T <- nrow(mXX)
  # eig <- eigs_sym(mXX, r, which = "LM", sigma = NULL, lower = TRUE)
  # meigvec <- eig$vectors 
  # mF <- sqrt(T) * meigvec       # estimated factors
  # Lambda = (t(mF) %*% SX)/T
  # SC = mF %*% Lambda
  # Xdm = scale(X,scale = FALSE)
  # stdX = t(Xdm) %*% Xdm / (T-1)
  # stdX = (diag(stdX))^0.5
  # stdX = diag(stdX)
  # C = SC %*% stdX
  obj <- svds(X,r) #nu = r,nv = r)
  C <- obj$u %*% diag(obj$d) %*% t(obj$v)  
  return(C)
}

###########################################################################

###########################################################################
# Part II: additional functions
# Part II 1) fdr: calculating fdr

fdr <- function(S, beta.true) {
  fdp = sum(beta.true[S] == 0)/max(1, length(S))
  return(fdp)
}
###########################################################################

###########################################################################
# Part II: additional functions
# Part II 2) pow: calculating power

pow <- function(S, beta.true) {
  tdp = sum(beta.true[S] != 0)/sum(beta.true != 0)
  return(tdp)
}

###########################################################################

###########################################################################
# n: number of observations
# p: number of variables
# r: number of factors
# ru: degree of correlation between variables
# theta: signal to noise ratio in factor models
# method: 1- normal: a n x p matrix X is generated form multivariate normal distribution
#         2- fac1: a n x p matrix X is generated form r factors and normal distribution error term
#         3- fac2: a n x p matrix X is generated form r factors and t distribution error term
DGPX <- function(n,p,ru,r,theta, method = "fac1"){
  if (method == "fac1") {
    Fa <- matrix(rnorm(n*r),n,r)
    Lam <- matrix(rnorm(r*p),r,p)
    e <- matrix(rnorm(n*p),n,p)
    X <- Fa %*% Lam + sqrt(theta)*e
    obj <- list(Factor = Fa, Lambda = Lam, X = X)
  } else if (method == "fac2") {
    Fa <- matrix(rnorm(n*r),n,r)
    Lam <- matrix(rnorm(r*p),r,p)
    e <- matrix(rnorm(n*p),n,p)
    v <- 8
    eta <- rchisq(p,v)
    eta <- diag((eta/(v-2))^-0.5)
    u <- e %*% eta 
    X <- Fa %*% Lam + sqrt(theta)*u
    obj <- list(Factor = Fa, Lambda = Lam, X = X)
  } else if (method == "normal") {
    Sigma <- diag(1,nrow = p, ncol = p)
    if (ru != 0) {
      for (i in 1:p) {
        for (j in 1:p) {
          Sigma[i,j] <- ru^(abs(i-j)) 
        }
      }
    }  
    mu <- rep(0,p)
    X <- mvrnorm(n, mu , Sigma)
    obj <- list(X = X)
  }
  return(obj)
}

###########################################################################

###########################################################################
# n: number of observations
# p: number of variables
# s: level of sparsity
# A: amplitude of signals
# method: 1- normal: a 1 x n vector y is generated form normal distribution
#         2- logistic:  a 1 x n vector y is generated form logistic distribution

DGPy <- function(X,s,A,c = 1, method = "linear", signal_to_noise = TRUE){
  if (method == "linear") {
    n = nrow(X)
    p = ncol(X)
    B.true <- rep(0,p)
    S.true <- sample(1:p, size = s) 
    S.true <- sort(S.true) # True Support
    C <- sample(c(A,-A), size = s , replace = TRUE) # coefficients for true signals
    B.true [S.true] <- C # A true vector of coefficients
    epsilon <- matrix(rnorm(n),n,1)  # Error Term
    y <- X %*% B.true + sqrt(c)*epsilon # Response variable
    if (signal_to_noise == TRUE) {
      yhat <- X %*% B.true
      sig_noise <- sum((yhat-mean(yhat))^2)/sum((y-mean(y))^2) # Signal to Noise Ratio
      obj = list(Beta = B.true, y = y,sig_noise = sig_noise )
    } else if (signal_to_noise == FALSE){
      obj = list(Beta = B.true, y = y)
    }
  }else if (method  == "nonlinear") {
    n = nrow(X)
    p = ncol(X)
    B.true <- rep(0,p)
    S.true <- sample(1:p, size = s) 
    S.true <- sort(S.true) # True Support
    C <- sample(c(A,-A), size = s , replace = T) # coefficients for true signals
    B.true [S.true] <- C # A true vector of coefficients
    epsilon <- matrix(rnorm(n),n,1)  # Error Term
    y <- 1/(1+exp(-(X %*% B.true))) + sqrt(c)*epsilon # Response variable
    if (signal_to_noise == TRUE) {
      yhat <- 1/(1+exp(-(X %*% B.true)))
      sig_noise <- sum((yhat-mean(yhat))^2)/sum((y-mean(y))^2) # Signal to Noise Ratio
      obj = list(Beta = B.true, y = y,sig_noise = sig_noise )
    } else if (signal_to_noise == FALSE) {
      obj = list(Beta = B.true, y = y)
    } 
  }
  return(obj)
}
##########################################################
## According to high-Knockoff paper, we first splite the sample into two disjoint group
sample.Splite = function(X, y){
  n   = nrow(X)
  p   = ncol(X)
  n0  = n/2
  n1  = n-n0
  full.sample   = c(1:n)
  scr.sample    = sample(full.sample,n0)
  sel.sample    = setdiff(full.sample,scr.sample)
  X_0           = X[scr.sample,]
  y_0           = y[scr.sample]
  X_1           = X[sel.sample,]
  y_1           = y[sel.sample]
  obj  = list(X0 = X_0, y0 = y_0, X1 = X_1, y1 = y_1)
  return(obj)
}

####################################################################

## In the screening step, we use the lasso to select a set
sel.set = function(X0, y0, n){
  n1 = n-nrow(X0)
  ## fit.lasso1 = scalreg(X0, y0, lam0 = NULL)     ## using the scalreg package
  ## set0  = order(abs(fit.lasso1$coefficients),decreasing=T)
  
  fit.lasso1 = cv.glmnet(X0, y0, dfmax = n/4)     ## using the glmnet package
  set0  = order(abs(coef(fit.lasso1)[-1,]),decreasing=T)
  scr.set      = set0[1:(n1/2)]
  scr.set.recy = set0[1:n1]
  obj = list(scr.set=scr.set, scr.set.recy=scr.set.recy)
  return(obj)
}

######################################################################
## After screening, we can get the selected variable or a reduced model
## y1 = X1_{\hat{S}_0}\beta^{p}+e1
sel.variable = function(scr.set, X0, X1, beta.true){
  X0.sel = X0[,scr.set]
  X1.sel = X1[,scr.set]
  beta.sel = beta.true[scr.set]
  obj  = list(X0.sel = X0.sel, X1.sel = X1.sel, beta.sel = beta.sel)
  return(obj)
}
# The end of the master R file
###########################################################################