# This include the high-dimensional Knockoff and Knockoff plus fliter to computer FDR and Power
# This master R file includes all the R functions for implementing the knockoff filter and 
# knockoff filter plus in high-dimensional linear regression.
#
# The IPAD framework has a range of potential applications involving precision 
# (innovated data) matrix including:
#   Linear and nonlinear classifications;
#   Dimension reduction; 
#   Portfolio management;
#   Feature screening and simultaneous inference.
#
# Written by 
# Yingying Fan (fanyingy@marshall.usc.edu) and Jinchi Lv (jinchilv@marshall.usc.edu), Mahrad Sharifvaghefi (sharifva@usc.edu), Yoshimasa Uematsu (yoshimasa.uematsu.e7@tohoku.ac.jp)
# USC, Nov. 2017
#
#
# List of R functions for ISEE:
# Part I: main functions
#   1) hdknockoff: main function of high-dimensional knockoff filter
#   2) hdknockoff_plus: main function of high-dimensional knockoff+
#   3) hdknockoffX: function generating the knockoff design matrix
# Part II: additional functions
#   1) fdr: calculating fdr
#   2) pow: calculating power

###########################################################################
# Part I: main functions
# Part I 1) hdknockoff: high-dimensional knockoff filter
# Input:
#   X:   n x p design matrix
#   y:   n x 1 response vector
#   q:   nominal FDR level
# Output: 
#   S:   set of discovered variables

forestknockoff <- function(X, y, q){
  
  n = nrow(X)
  p = ncol(X)/2
  
  # generate the knockoff design matrix
  
  obj1 = randomForest(X,y,importance = TRUE)
  betap.stan = obj1$importance[,1]
  
  # Calculate knockoff test statistics W_j
  W = abs(betap.stan[1:p]) - abs(betap.stan[(p+1):(2*p)])
  
  # find the knockoff threshold T
  t = sort(c(0, abs(W)))
  ratio = c(rep(0, p))
  for (j in 1:p) {
    ratio[j] = sum(W <= -t[j])/max(1, sum(W >= t[j]))
  }
  id = which(ratio <= q)[1]
  if(length(id) == 0){
    T = Inf
  } else {
    T = t[id]
  }
  
  # set of discovered variables
  S = which(W >= T)
  
  return(S)
}
###########################################################################

###########################################################################
# Part I: main functions
# Part I 1) hdknockoff plus: high-dimensional knockoff plus filter
# Input:
#   X:   n x p design matrix
#   y:   n x 1 response vector
#   q:   nominal FDR level
# Output: 
#   S_plus:   set of discovered variables

forestknockoff_plus <- function(X, y, q){
  
  n = nrow(X)
  p = ncol(X)/2

  obj1 = randomForest(X,y,importance = TRUE)
  betap.stan = obj1$importance[,1]
  
  # Calculate knockoff test statistics W_j
  W = abs(betap.stan[1:p]) - abs(betap.stan[(p+1):(2*p)])

  # find the knockoff threshold T
  t = sort(c(0, abs(W)))
  ratio = c(rep(0, p))
  for (j in 1:p) {
    ratio[j] = (1+sum(W <= -t[j]))/max(1, sum(W >= t[j]))
  }
  id = which(ratio <= q)[1]
  if(length(id) == 0){
    T = Inf
  } else {
    T = t[id]
  }
  
  # set of discovered variables
  S_plus = which(W >= T)
  
  return(S_plus)
}
###########################################################################

###########################################################################
# Part I: main functions
# Part I 2) hdknockoffX: function generating the knockoff design matrix
# Input:
#   X:      n x p design matrix
#   y:      n x 1 response vector
# Output: 
#   Xnew:   n x (2p) design matrix = [X X_ko]
# Methods: 1- facpermu: Knockoff variables are constructed by using factor model with derangement premutation of residuals
#          2- facrand: Knockoff variables are constructed by using factor model with random error drown form unit sphere 
#          3- linsparse: Knockoff variables are constructed by using linear sparse model
#          4- BCknockoff1: Barber and Candes equicorrelated method for construction of knockoff variables
#          5- hdBCknockoff1: Barber and Candes equicorrelated method for construction of knockoff variables in high dimensional setting
hdknockoffX <- function(X,y,r,method = "facrand", lambda = "CV"){
  if (method == "facrand") {
    n = nrow(X)
    p = ncol(X)
    C = fFact(X,r)
    res = X - C
    resScale = sqrt(apply(res^2,2,sum))
    resScale = diag(resScale)
    e = matrix(rnorm(n*p),n,p)
    eScale = sqrt(apply(e^2,2,sum))
    eScale = diag(eScale^-1)
    nres = e %*% resScale %*% eScale
    X_ko = C + nres
    Xnew = cbind(X, X_ko)
  } else if (method == "BCknockoff1") {
    X_ko <- knockoff.create(X,method = "equicorrelated")
    Xnew = cbind(X, X_ko)
  } else if (method == "hdBCknockoff1") {
    n = nrow(X)
    p = ncol(X)
    y0 = y[1:round(n/2)]
    X0 = X[1:round(n/2),]
    y1 = y[(round(n/2)+1):n]
    X1 = X[(round(n/2)+1):n,]
    # Finding Optimal value of Lambda
    Op_lambda <- cv.glmnet(X0,y0)$lambda.min
    # apply glmnet to find beta vector under original scale
    Beta <- glmnet(X0,y0,lambda = Op_lambda)$beta
    Beta <- as.vector(Beta)
    X_ko <- knockoff.create(X1[,Beta > 0],method = "equicorrelated")
    X_ko <- rbind(X0[,Beta > 0],X_ko)
    Xnew = cbind(X[,Beta > 0], X_ko)
  }
  return(Xnew)
}
###########################################################################

###########################################################################
fFact <- function(X, r){
  obj <- svds(X,r)
  C <- obj$u %*% diag(obj$d) %*% t(obj$v)  
  return(C)
}

###########################################################################

###########################################################################
# Part II: additional functions
# Part II 1) fdr: calculating fdr

fdr <- function(S, beta.true) {
  fdp = sum(beta.true[S] == 0)/max(1, length(S))
  return(fdp)
}
###########################################################################

###########################################################################
# Part II: additional functions
# Part II 2) pow: calculating power

pow <- function(S, beta.true) {
  tdp = sum(beta.true[S] != 0)/sum(beta.true != 0)
  return(tdp)
}

###########################################################################

###########################################################################
# n: number of observations
# p: number of variables
# r: number of factors
# ru: degree of correlation between variables
# theta: signal to noise ratio in factor models
# method: 1- normal: a n x p matrix X is generated form multivariate normal distribution
#         2- fac1: a n x p matrix X is generated form r factors and normal distribution error term
#         3- fac2: a n x p matrix X is generated form r factors and t distribution error term
DGPX <- function(n,p,ru,r,theta, method = "fac1"){
  if (method == "fac1") {
    Fa <- matrix(rnorm(n*r),n,r)
    Lam <- matrix(rnorm(r*p),r,p)
    e <- matrix(rnorm(n*p),n,p)
    X <- Fa %*% Lam + sqrt(theta)*e
    obj <- list(Factor = Fa, Lambda = Lam, X = X)
  } else if (method == "fac2") {
    Fa <- matrix(rnorm(n*r),n,r)
    Lam <- matrix(rnorm(r*p),r,p)
    e <- matrix(rnorm(n*p),n,p)
    v <- 8
    eta <- rchisq(p,v)
    eta <- diag((eta/(v-2))^-0.5)
    u <- e %*% eta 
    X <- Fa %*% Lam + sqrt(theta)*u
    obj <- list(Factor = Fa, Lambda = Lam, X = X)
  } else if (method == "normal") {
    Sigma <- diag(1,nrow = p, ncol = p)
    if (ru != 0) {
      for (i in 1:p) {
        for (j in 1:p) {
          Sigma[i,j] <- ru^(abs(i-j)) 
        }
      }
    }  
    mu <- rep(0,p)
    X <- mvrnorm(n, mu , Sigma)
    obj <- list(X = X)
  }
  return(obj)
}
###########################################################################

###########################################################################
# n: number of observations
# p: number of variables
# s: level of sparsity
# A: amplitude of signals
# method: 1- normal: a 1 x n vector y is generated form normal distribution
#         2- logistic:  a 1 x n vector y is generated form logistic distribution

DGPy <- function(X,s,A,c = 1, method = "linear"){
  if (method == "linear") {
    n = nrow(X)
    p = ncol(X)
    B.true <- rep(0,p)
    S.true <- sample(1:p, size = s) 
    S.true <- sort(S.true) # True Support
    C <- sample(c(A,-A), size = s , replace = T) # coefficients for true signals
    B.true [S.true] <- C # A true vector of coefficients
    epsilon <- matrix(rnorm(n),n,1)  # Error Term
    y <- X %*% B.true + sqrt(c)*epsilon # Response variable
    obj = list(Beta = B.true, y = y) 
  }else if (method  == "nonlinear1") {
    n = nrow(X)
    p = ncol(X)
    B.true <- rep(0,p)
    S.true <- sample(1:p, size = s) 
    S.true <- sort(S.true) # True Support
    C <- sample(c(A,-A), size = s , replace = T) # coefficients for true signals
    B.true [S.true] <- C # A true vector of coefficients
    epsilon <- matrix(rnorm(n),n,1)  # Error Term
    y <- (X %*% B.true)^3 + sqrt(c)*epsilon # Response variable
    obj = list(Beta = B.true, y = y) 
  }else if (method  == "nonlinear2") {
    n = nrow(X)
    p = ncol(X)
    B.true <- rep(0,p)
    S.true <- sample(1:p, size = s) 
    S.true <- sort(S.true) # True Support
    C <- sample(c(A,-A), size = s , replace = T) # coefficients for true signals
    B.true [S.true] <- C # A true vector of coefficients
    epsilon <- matrix(rnorm(n),n,1)  # Error Term
    y <-   sin(X %*% B.true)*exp((X %*% B.true)) + sqrt(c)*epsilon # Response variable
    obj = list(Beta = B.true, y = y) 
  }
  return(obj)
}
# The end of the master R file
###########################################################################